use core::{
    mem,
    sync::atomic::{AtomicU32, Ordering},
};

const LOCKED_BIT: u32 = 0b001;
const EMPTY: u32 = 0;

/// A simple lock implementation using an atomic u32.
#[repr(transparent)]
pub struct Lock {
    inner: AtomicU32,
}

impl Lock {
    /// Creates a new lock in the unlocked state.
    pub const fn new() -> Self {
        Self {
            inner: AtomicU32::new(0),
        }
    }

    pub fn as_ptr(&self) -> *mut u32 {
        self.inner.as_ptr()
    }

    pub unsafe fn from_ptr<'a>(ptr: *mut u32) -> &'a Self {
        // SAFETY: The caller must ensure that `ptr` is not aliased, and lasts
        // for the lifetime of the `Lock`.
        unsafe { mem::transmute(AtomicU32::from_ptr(ptr)) }
    }

    /// Acquires the lock, blocking until it is available.
    pub fn lock(&self) {
        // attempt acquiring the lock with no contention.
        if self
            .inner
            .compare_exchange_weak(EMPTY, LOCKED_BIT, Ordering::Acquire, Ordering::Relaxed)
            .is_ok()
        {
            // We successfully acquired the lock.
            return;
        } else {
            self.lock_slow();
        }
    }

    pub fn unlock(&self) {
        // use release semantics to ensure that all previous writes are
        // available to other threads.
        self.inner.fetch_and(!LOCKED_BIT, Ordering::Release);
    }

    fn lock_slow(&self) {
        // The lock is either locked, or someone is waiting for it:

        let mut spin_wait = SpinWait::new();
        let mut state = self.inner.load(Ordering::Acquire);
        loop {
            // If the lock isn't locked, we can try to acquire it.
            if state & LOCKED_BIT == 0 {
                // Try to acquire the lock.
                match self.inner.compare_exchange_weak(
                    state,
                    state | LOCKED_BIT,
                    Ordering::Acquire,
                    Ordering::Relaxed,
                ) {
                    Ok(_) => {
                        // We successfully acquired the lock.
                        return;
                    }
                    Err(new_state) => {
                        // The state changed, we need to check again.
                        state = new_state;
                        continue;
                    }
                }
            }

            if {
                let spun: bool;
                #[cfg(feature = "std")]
                {
                    spun = spin_wait.spin_yield();
                }
                #[cfg(not(feature = "std"))]
                {
                    spun = spin_wait.spin();
                }

                spun
            } {
                // We can spin for a little while and see if it becomes available.
                state = self.inner.load(Ordering::Relaxed);
                continue;
            }

            // If we reach here, we need to park the thread.
            atomic_wait::wait(&self.inner, LOCKED_BIT);

            if self
                .inner
                .compare_exchange_weak(
                    state,
                    state | LOCKED_BIT,
                    Ordering::Acquire,
                    Ordering::Relaxed,
                )
                .is_ok()
            {
                // We successfully acquired the lock after being woken up.
                return;
            }

            spin_wait.reset();
            state = self.inner.load(Ordering::Relaxed);
        }
    }

    pub fn wait(&self) {
        let state = self.inner.load(Ordering::Acquire);
        atomic_wait::wait(&self.inner, state);
    }

    pub fn wake_one(&self) {
        // Notify one thread waiting on this lock.
        atomic_wait::wake_one(&self.inner);
    }
}

// from parking_lot_core
pub struct SpinWait {
    counter: u32,
}

impl SpinWait {
    /// Creates a new `SpinWait` with an initial counter value.
    pub const fn new() -> Self {
        Self { counter: 0 }
    }

    /// Resets the counter to zero.
    pub fn reset(&mut self) {
        self.counter = 0;
    }

    pub fn spin(&mut self) -> bool {
        if self.counter >= 10 {
            // If the counter is too high, we signal the caller to potentially park.
            return false;
        }
        self.counter += 1;

        // spin for a small number of iterations based on the counter value.
        for _ in 0..(1 << self.counter) {
            core::hint::spin_loop();
        }

        true
    }

    #[cfg(feature = "std")]
    pub fn spin_yield(&mut self) -> bool {
        if self.counter >= 10 {
            // If the counter is too high, we signal the caller to potentially park.
            return false;
        }
        self.counter += 1;

        if self.counter >= 3 {
            // spin for a small number of iterations based on the counter value.
            for _ in 0..(1 << self.counter) {
                core::hint::spin_loop();
            }
        } else {
            // yield the thread and wait for the OS to reschedule us.
            std::thread::yield_now();
        }

        true
    }
}

// taken from `std`
#[derive(Debug)]
#[repr(transparent)]
pub struct Parker {
    mutex: AtomicU32,
}

impl Parker {
    const PARKED: u32 = u32::MAX;
    const EMPTY: u32 = 0;
    const NOTIFIED: u32 = 1;
    pub fn new() -> Self {
        Self {
            mutex: AtomicU32::new(Self::EMPTY),
        }
    }

    pub fn as_ptr(&self) -> *mut u32 {
        self.mutex.as_ptr()
    }

    pub unsafe fn from_ptr<'a>(ptr: *mut u32) -> &'a Self {
        // SAFETY: The caller must ensure that `ptr` is not aliased, and lasts
        // for the lifetime of the `Parker`.
        unsafe { mem::transmute(AtomicU32::from_ptr(ptr)) }
    }

    pub fn is_parked(&self) -> bool {
        self.mutex.load(Ordering::Acquire) == Self::PARKED
    }

    pub fn park(&self) {
        self.park_inner(|| ());
    }

    pub fn park_with_callback<F>(&self, before_sleep: F)
    where
        F: FnOnce(),
    {
        // This function is called when the thread is about to park.
        self.park_inner(before_sleep);
    }

    // If the caller wants to park the thread on a mutex'd condition, it is
    // possible for a deadlock to occur when the mutex is dropped just before
    // this parker atomically changes it's state to `PARKED`. For that reason,
    // we use a callback to allow the caller to perform any necessary actions
    // before parking, but after the parker is set to `PARKED`. If another
    // thread then checks the condition and checks if this thread should be
    // woken, we will be immediately notified.
    // Thusly, it is important that any caller synchronise any conditionals with
    // a mutex externally, and then unlock that mutex in `before_sleep`.
    fn park_inner<F>(&self, before_sleep: F)
    where
        F: FnOnce(),
    {
        if self.mutex.fetch_sub(1, Ordering::Acquire) == Self::NOTIFIED {
            before_sleep();
            // The thread was notified, so we can return immediately.
            return;
        }

        before_sleep();

        loop {
            atomic_wait::wait(&self.mutex, Self::PARKED);

            // We check whether we were notified or woke up spuriously with
            // acquire ordering in order to make-visible any writes made by the
            // thread that notified us.
            if self.mutex.swap(Self::EMPTY, Ordering::Acquire) == Self::NOTIFIED {
                // The thread was notified, so we can return immediately.
                return;
            } else {
                // spurious wakeup, so we need to re-park.
                continue;
            }
        }
    }

    pub fn unpark(&self) {
        // write with Release ordering to ensure that any writes made by this
        // thread are made-available to the unparked thread.
        if self.mutex.swap(Self::NOTIFIED, Ordering::Release) == Self::PARKED {
            // The thread was parked, so we need to notify it.
            atomic_wait::wake_one(&self.mutex);
        } else {
            // The thread was not parked, so we don't need to do anything.
        }
    }
}
